#! /usr/local/python2.4

import itertools
import operator

import helper


INF = float('inf')

totalInactive = 0

class Marking(object):

  def __init__(self, seq):
    self.tuple = tuple(seq)

  def __str__(self):
      return str(self.tuple)
  @helper.Counter.count
  def Accelerate(self, highMark):
    """Return the accelerated marker.

    Args:
      highMark: A marker that must dominate self.
    """
    assert highMark > self
    def acc(high, low):
      return {True: INF, False: high}[high > low]

    tuples = itertools.izip(highMark.tuple, self.tuple)
    return Marking(acc(high, low) for high, low in tuples)

  def AddTuple(self, tuple):
    return Marking(map(operator.add, self.tuple, tuple))

  @property
  def isNonNegative(self):
    return self >= Marking((0,) * len(self))

  def __len__(self):
    return len(self.tuple)

  @helper.Counter.count
  def __le__(self, other):
    assert isinstance(other, Marking)

    pairs = ((it1, it2) for it1, it2 in itertools.izip(self.tuple, other.tuple) if it1 > it2)
    for _ in pairs:
      return False
    return True

  def __lt__(self, other):
    assert isinstance(other, Marking)
    return self != other and self <= other

  def __gt__(self, other):
    assert isinstance(other, Marking)
    return other < self

  def __ge__(self, other):
    assert isinstance(other, Marking)
    return other <= self

  def __eq__(self, other):
    return self.tuple == other.tuple

  def __ne__(self, other):
    return self.tuple != other.tuple

  def __hash__(self):
    return self.tuple.__hash__()

def GetUniformTransition(deltas, tmins=None):
  """delta is a tuple. tmin are Marking."""
  @helper.Counter.count
  def Transition(mark):
    assert mark is None or isinstance(mark, Marking)
    if tmins is None:
      valid_deltas = deltas
    else:
      valid_deltas = (d for d, min in zip(deltas, tmins) if mark >= min)
    canditates = (mark.AddTuple(d) for d in valid_deltas)

    return [c for c in canditates if c.isNonNegative]

  return Transition



class Node(object):

  def __init__(self, mark, parent):
    """Initialize a Node.

    Args:
      mark: the value of the Node to create.
      parent. A Node representing the parent of the Node to create.
    """
    assert mark is None or isinstance(mark, Marking)
    self.value = mark
    self.parent = parent
    self.children = set()
    self.min_ancestors = set()
    self.dead = False

  def __str__(self):
    return "%s-%s%s" % (self.value, id(self), {True: '-X', False: ''}[self.dead])

  def PrintSubTree(self, indent=0, step=4):
    print indent*'-', self
    [c.PrintSubTree(indent+step) for c in self.children]

  def IsComparableWith(self, other):
    return  self.value <= other.value or self.value > other.value

  @helper.Counter.count
  def AddChild(self, node):
    assert node.parent == self
    self.children.add(node)

  @helper.Counter.count
  def DettachFromParent(self):
    if self.parent is not None: 
      self.parent.children.remove(self)

  @helper.Counter.count
  def AddMinAncestor(self, mark):
    """Add a descendant only if it is minimal and removed Node that dominates mark."""
    assert isinstance(mark, Marking)
    toRemove = set()
    for d in self.min_ancestors:
      if d <= mark:
        return
      if mark < d:
        toRemove.add(d)
    self.min_ancestors.difference_update(toRemove)
    self.min_ancestors.add(mark)

  @helper.Counter.count
  def IsMinDominated(self, mark):
    """Return True if either value or one of min_descendant is dominated by mark."""
    ## TODO: add value in min set: better when value dominates some min in the set.
    assert isinstance(mark, Marking)
    assert mark != self.value # TODO Remove assert.
    if mark > self.value:
        return True
    min_dom = (anc for anc in self.min_ancestors if mark > anc)
    for _ in min_dom:
      print '-'*5, 'is min dom'
      return True
    return False

#    for anc in self.min_ancestors:
#      assert mark != anc
#      if mark > anc:
#        return True
#    return mark > self.value

  def _PathToRoot(self):
    """Return the path from self to root."""
    current_node = self
    while current_node != None:
      yield current_node
      current_node = current_node.parent

  @property
  def pathFromRoot(self):
    return reversed(list(self._PathToRoot()))


  @property
  def siblings(self):
    return (node for node in self.parent.children if node != self)

class CoverSet(object):

  def __init__(self, transition, initial):
    """Initialize a CoverSet.

    Args:
      transition: A function from tuple to list of tuple.
      initial: A Marker.
    """
    self.transition = transition
    self.root = Node(initial, None)
    self.candidates = []
    self._AddCandidates(self.root)

  def ToSet(self):
    return set(map(lambda n: n.value.tuple, Util.Descendants([self.root])))


  def _AddNewCandidate(self, new_node):
      nodes = (Node(m, new_node) for m in self.transition(new_node.value))
      self.candidates.extend(nodes)


  @staticmethod
  def _ProcessToKill(not_ancestors, new_node):
    global totalInactive
    desc = Util.Descendants(not_ancestors)
    dominated = itertools.ifilter(lambda v:v.IsMinDominated(new_node.value), desc)
    # new_node has children only when acc != None
    toKill = itertools.chain(dominated, new_node.children)

    for node in toKill: 
      node.DettachFromParent()
      for d in Util.Descendants([node]):
        if d.dead == False:
          totalInactive = totalInactive +1
          d.dead = True


  @staticmethod
  def _AccelerateNode(next_node, acc):
      """ computes the accelrated marking, set the next_node.parent to acc.parent."""
      acc_mark = acc.value.Accelerate(next_node.value)
      next_node.value = acc_mark
      CoverSet._AddMinimumAncestors(next_node,acc)
      next_node.parent = acc.parent
  @staticmethod
  def _AddMinimumAncestors(high_node, low_node):
      parent=high_node.parent
      while parent is not None and parent != low_node.parent:
          high_node.AddMinAncestor(parent.value)
          parent = parent.parent


  @staticmethod
  def _ProcessAncestors(root, node):
    """Seek domination in ancestors and replace if domination found.
    returns acc if next_node dominates its ancestor acc: it will be accelerated
    returns the set not_ancestor: contains candidates to be killed
    (siblings of ancestors and acc if not None)"""
    not_ancestors = []

    path = node.pathFromRoot
    path.next()

    hasInComparableParent = lambda anc: not node.IsComparableWith(anc.parent)
    treatSiblings = itertools.takewhile(hasInComparableParent, path)

    currentNode = root # if none are incomparable, we need to accelerate root.
    for currentNode in treatSiblings:
      not_ancestors.extend(currentNode.siblings)

    hasAcc = currentNode != node
    acc = None
    if hasAcc:
      acc = currentNode
      not_ancestors.append(acc) # Acc will be replaced by node and thus has to be killed later.

    return acc, not_ancestors

  def _CheckDomination(self, next_node):
    """Check whether next_node is dominated by a node in the tree."""
    # We could other strategy for poping up from remainings.
    desc = Util.Descendants([self.root])
    dominated = itertools.ifilter(lambda n: n.value >= next_node.value, desc) # TODO
    for _ in dominated:
      return True
    return False

  def _AddCandidates(self, node):
    self.candidates.extend(Node(n, node) for n in self.transition(node.value))

  def ProcessNext(self, next_node):

    assert next_node.parent.dead == False
    assert next_node.parent is not None

    acc, not_ancestors = CoverSet._ProcessAncestors(self.root, next_node)

    if acc is not None:
      CoverSet._AccelerateNode(next_node, acc)

    if acc == self.root: 
      self.root = next_node
    else:
      next_node.parent.AddChild(next_node)

    CoverSet._ProcessToKill(not_ancestors, next_node)

    self._AddNewCandidate(next_node)

  @helper.profile
  def Compute(self):
    """Minimal Coverability Set Computation."""
    global totalInactive
    totalInactive = 0  # Number of deactivated nodes
    verbose = 1        # To enable displays
    totalNodes1 = 1    # Number of nodes popped from the Waiting list
    totalNodes2 = 1    # Number of nodes that are processed
    mode = 0           # Defines DFS or BFS
    if (mode and verbose):
      print "Depth-First Search"
    elif (verbose):
      print "Breadth-First Search"
    while len(self.candidates) !=0:
      if (mode==1):
        candidate = self.candidates.pop() # DFS
      else:
        candidate = self.candidates.pop(0) # BFS
      totalNodes1 = totalNodes1+1
      if not candidate.dead and candidate.parent.dead == False:
        totalNodes2 = totalNodes2+1
        if not self._CheckDomination(candidate):
          self.ProcessNext(candidate)
        else:
          totalInactive = totalInactive+1
    if (verbose):
      print "totals : ",totalNodes1,"/",totalNodes2,"/",totalInactive


  @helper.profile
  def KM_Compute(self):
    """Karp And Miller Procedure."""
    verbose = 0        # To enable displays
    totalNodes1 = 0    # 
    totalNodes2 = 0
    while len(self.candidates) !=0:
      candidate = self.candidates.pop()
      totalNodes1 = totalNodes1+1
      if not candidate.KM_dominated():
        candidate.KM_accelerate()
        self._AddNewCandidate(candidate)
        totalNodes2 = totalNodes2+1
    if (verbose):
      print "totals : ",totalNodes1,"/",totalNodes2
      print "tree size: ",self.SizeOfTree()



class Util(object):

  @staticmethod
  def Pairwise(it):
        """Return each element and its successer."""
        a, b = itertools.tee(it)
        b.next()
        return itertools.izip(a,b)

  @staticmethod
  def Descendants(subTrees): 
    while len(subTrees) !=0:
        node = subTrees.pop()
        assert not node.dead
        yield node
        subTrees.extend(c for c in node.children if not c.dead)



